from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.task import Task
from sklearn.ensemble import RandomForestRegressor
import numpy as np
import pickle as pkl
import os


def train_one_oracle(target_assay, x_i, y_i):
    """Train a Random Forest Regression Tree using scikit-learn
    and save that classifier to the disk

    Args:

    x: np.ndarray
        the training features for the decision tree represented as a
        matrix with a shape like [n_samples, n_features]
    y: np.ndarray
        the training labels for the decision tree represented as a
        matrix with a shape like [n_samples, 1]
    """

    est = RandomForestRegressor(
        n_estimators=500,
        criterion="mse",
        max_depth=32,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.,
        max_features="auto",
        max_leaf_nodes=None,
        min_impurity_decrease=0.,
        min_impurity_split=None,
        bootstrap=True,
        oob_score=False,
        n_jobs=24,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None)

    est.fit(x_i, y_i[:, 0])
    os.makedirs(os.path.join(
        DATA_DIR, f'molecule_activity_v0'), exist_ok=True)
    with open(os.path.join(
            DATA_DIR, f'molecule_activity_v0/'
                      f'rfr_{target_assay}.pkl'), 'wb') as f:
        pkl.dump(est, f)
    r2 = est.score(x_i, y_i[:, 0])
    print(r2)


class MoleculeActivityV0Task(Task):

    def score(self, x):
        return NotImplemented

    def __init__(self,
                 target_assay=600885,
                 split_percentile=80,
                 ys_noise=0.0):  # this choice has the most spread
        """Create a task for designing super conducting materials that
        have a high critical temperature

        Support target Assay, Data set Size
        688150,               2k
        600886,               5k
        600885,               5k
        688537,               3k
        688597,               3k

        Args:

        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        maybe_download('1_8c7uln7vzLbMmoviJhGWRqy-OyMZovc',
                       os.path.join(DATA_DIR, 'molecule_activity_X.npy'))
        maybe_download('1nZgMkKs6_hKb8o1Oy-KhGBLtI7HgqD2K',
                       os.path.join(DATA_DIR, 'molecule_activity_c.npy'))
        maybe_download('1vDycVRmfHi_-NiMjDaFrs-M4r9hZKr8T',
                       os.path.join(DATA_DIR, 'molecule_activity_y.npy'))
        maybe_download('1TFpIiLlpOYUoRRL3vexIhuY2Z_ZuAyxa',
                       os.path.join(DATA_DIR, 'molecule_activity_v0.zip'))

        x = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_X.npy')).astype(np.float32)
        c = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_c.npy')).astype(np.int32)
        y = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_y.npy')).astype(np.float32)

        # select the examples for this assay
        matches = np.where(np.equal(c, target_assay))[0]
        x = x[matches]
        y = y[matches]
        x = np.stack([1.0 - x, x], axis=2)

        # remove all samples above the qth percentile in the data set
        split_temp = np.percentile(y[:, 0], split_percentile)
        indices = np.where(y <= split_temp)[0]
        x = x[indices].astype(np.float32)
        y = y[indices].astype(np.float32)

        mean_y = np.mean(y, axis=0, keepdims=True)
        st_y = np.std(y - mean_y, axis=0, keepdims=True)
        y = y + np.random.normal(0.0, 1.0, y.shape) * st_y * ys_noise
        self.x = x
        self.y = y

        # load the oracle for this assay
        with open(os.path.join(
                DATA_DIR, f'molecule_activity_v0/'
                          f'rfr_{target_assay}.pkl'), 'rb') as f:
            self.oracle = pkl.load(f)
        self.score = np.vectorize(
            self.scalar_score, signature='(n,2)->(1)')

    def scalar_score(self,
                     x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        return self.oracle.predict(
            (x[np.newaxis, :, 1] > 0.5).astype(np.float32))



"""Design for the registry is inspired significantly by:
https://github.com/openai/gym/blob/master/gym/envs/registration.py
"""

import re
import importlib


# these are the default task name match criterion and error messages
TASK_PATTERN = re.compile(r'([\w:.-]+)-v(\d+)$')
MISMATCH_MESSAGE = 'Attempted to register malformed task name: {}. (' \
                   'Currently all names must be of the form {}.)'
DEPRECATED_MESSAGE = 'Task {} not found (valid versions include {})'
UNKNOWN_MESSAGE = 'No registered task with name: {}'
REREGISTRATION_MESSAGE = 'Cannot re-register id: {}'


# this is used to import data set classes dynamically
def import_name(name):
    mod_name, attr_name = name.split(":")
    return getattr(importlib.import_module(mod_name), attr_name)


class TaskSpecification(object):

    def __init__(self,
                 name,
                 entry_point,
                 kwargs=None):
        """Create a specification for a model-based optimization task that
        dynamically imports that task when self.make is called.

        Args:

        name: str
            the name of the model-based optimization task which must match
            with the regex expression given by TASK_PATTERN
        entry_point: str or callable
            the import path to the target task class or a callable that
            returns the target task class when called
        kwargs: dict
            additional keyword arguments that are provided to the task
            class when it is initialized for the first time
        """

        # store the init arguments
        self.name = name
        self.entry_point = entry_point
        self.kwargs = {} if kwargs is None else kwargs

        # check if the name matches with the regex template
        match = TASK_PATTERN.search(name)

        # if there is no match raise a ValueError
        if not match:
            raise ValueError(
                MISMATCH_MESSAGE.format(name, TASK_PATTERN.pattern))

        # otherwise select the task name from the regex match
        self.task_name = match.group(0)

    def make(self, **additional_kwargs):
        """Instantiates the intended task using the additional
        keyword arguments provided in this function.

        Args:

        additional_kwargs: dict
            additional keyword arguments that are provided to the task
            class when it is initialized for the first time

        Returns:

        task: Task
            an instantiation of the task specified by task name which
            conforms to the standard task api
        """

        # use additional_kwargs to override self.kwargs
        kwargs = self.kwargs.copy()
        kwargs.update(additional_kwargs)

        # if self.entry_point is a function call it
        if callable(self.entry_point):
            return self.entry_point(**kwargs)

        # if self.entry_point is a string import it first
        elif isinstance(self.entry_point, str):
            return import_name(self.entry_point)(**kwargs)

    def __repr__(self):
        return "TaskSpecification({}, {})".format(self.name, self.entry_point)


class TaskRegistry(object):

    def __init__(self):
        """Provide a global interface for registering model-based
        optimization tasks that remain stable over time
        """

        self.task_specs = {}

    def make(self, name, **kwargs):
        """Instantiates the intended task using the additional
        keyword arguments provided in this function.

        Args:

        name: str
            the name of the model-based optimization task which must match
            with the regex expression given by TASK_PATTERN
        additional_kwargs: dict
            additional keyword arguments that are provided to the task
            class when it is initialized for the first time

        Returns:

        task: Task
            an instantiation of the task specified by task name which
            conforms to the standard task api
        """

        return self.spec(name).make(**kwargs)

    def all(self):
        """Generate a list of the names of all currently registered
        model-based optimization tasks

        Returns:

        names: list
            a list of names that corresponds to all currently registered
            tasks where names[i] is suitable for self.make(names[i])
        """

        return self.task_specs.values()

    def spec(self, name):
        """Looks up the task specification identifed by the task
        name argument provided here

        Args:

        name: str
            the name of the model-based optimization task which must match
            with the regex expression given by TASK_PATTERN

        Returns:

        spec: TaskSpecification
            a specification whose make function will dynamically import
            and create the task specified by 'name'
        """

        # check if the name matches with the regex template
        match = TASK_PATTERN.search(name)

        # if there is no match raise a ValueError
        if not match:
            raise ValueError(
                MISMATCH_MESSAGE.format(name, TASK_PATTERN.pattern))

        # try to locate the task specification
        try:
            return self.task_specs[name]

        # if it does not exist try to find out why
        except KeyError:

            # make a list of all similar registered tasks
            task_name = match.group(0)
            matching_tasks = [valid_name for valid_name, valid_spec
                              in self.task_specs.items()
                              if task_name == valid_spec.task_name]

            # there is a similar matching task
            if matching_tasks:
                raise ValueError(
                    DEPRECATED_MESSAGE.format(name, matching_tasks))

            # there are no similar matching tasks
            else:
                raise ValueError(
                    UNKNOWN_MESSAGE.format(name))

    def register(self, name, entry_point, **kwargs):
        """Register a specification for a model-based optimization task that
        dynamically imports that task when self.make is called.

        Args:

        name: str
            the name of the model-based optimization task which must match
            with the regex expression given by TASK_PATTERN
        entry_point: str or callable
            the import path to the target task class or a callable that
            returns the target task class when called
        kwargs: dict
            additional keyword arguments that are provided to the task
            class when it is initialized for the first time
        """

        # raise an error if that task is already registered
        if name in self.task_specs:
            raise ValueError(REREGISTRATION_MESSAGE.format(name))

        # otherwise add that task to the collection
        self.task_specs[name] = TaskSpecification(name, entry_point, **kwargs)


# create a global task registry
registry = TaskRegistry()


# wrap the TaskRegistry.register function globally
def register(name, entry_point, **kwargs):
    return registry.register(name, entry_point, **kwargs)


# wrap the TaskRegistry.make function globally
def make(name, **kwargs):
    return registry.make(name, **kwargs)


# wrap the TaskRegistry.spec function globally
def spec(name):
    return registry.spec(name)



from design_bench.task import Task
import numpy as np


class QuadraticTask(Task):

    def __init__(self,
                 global_optimum=(-1.0, 4.0, 2.0, -3.0, 5.0, 1.0,),
                 oracle_noise_std=0.2,
                 dataset_size=100,
                 split_percentile=80,
                 ys_noise=0.0):
        """Create a toy Gaussian Process optimization task using
        the provided sampling bounds

        Args:

        global_optimum: tuple or list
            the maximal location of the quadratic function whose length
            dictates the number of channels in the task dataset
        oracle_noise_std: float
            the standard deviation of the noise added to the function
            used to score the sampled x values
        dataset_size: int
            the number of initial samples to populate the dataset with
            must be a positive integer
        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        global_optimum = np.array(global_optimum)[np.newaxis]
        self.global_optimum = global_optimum.astype(np.float32)
        self.oracle_noise_std = oracle_noise_std

        z = np.random.randn(dataset_size, len(global_optimum))
        x = (self.global_optimum + z).astype(np.float32)
        y = self.score(x).astype(np.float32)

        split_value = np.percentile(y[:, 0], split_percentile)
        indices = np.where(y <= split_value)[0]
        y = y[indices]
        x = x[indices]
        self.x = x
        self.y = y

        mean_y = np.mean(self.y, axis=0, keepdims=True)
        st_y = np.std(self.y - mean_y, axis=0, keepdims=True)
        self.y = self.y + np.random.normal(0.0, 1.0, self.y.shape) * st_y * ys_noise

    def score(self,
              x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        x = x - self.global_optimum
        z = self.oracle_noise_std * np.random.randn(x.shape[0], 1)
        return -np.sum(x ** 2, axis=1, keepdims=True) + z + 6.0



from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.task import Task
from tape import TAPETokenizer
from tape import ProteinBertForValuePrediction
import torch
import pandas as pd
import numpy as np
import os


def onehottify(x, n=None, dtype=float):
    """1-hot encode x with the max value n (computed from data if n is None)."""
    x = np.asarray(x)
    n = np.max(x) + 1 if n is None else n
    return np.eye(n, dtype=dtype)[x]


class GFPV1Task(Task):

    def __init__(self,
                 split_percentile=20,
                 internal_batch_size=1,
                 use_cuda=True,
                 ys_noise=0.0):
        """Load the GFP data set which includes maps from discrete
        protein designs to fluorescence scores

        Args:

        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        maybe_download('1_jcPkQ-M1FRhkEONoE57WEbp_Rivkho2',
                       os.path.join(DATA_DIR, 'gfp_data.csv'))
        maybe_download('1R2UaplzHjMaWwsu-bT-kcOAx7GS1HYRu',
                       os.path.join(DATA_DIR, 'gfp_transformer_pretrained.zip'))
        self.batch_size = internal_batch_size
        self.use_cuda = use_cuda

        # load the static dataset
        df = pd.read_csv(os.path.join(DATA_DIR, 'gfp_data.csv'))

        # remove all proteins with fluorescence below the mean
        df = df.loc[df.loc[(df['medianBrightness'] >
                            df['medianBrightness'].mean())].index]

        # remove all proteins with a stop marker
        df = df.loc[df.loc[
            ~df['aaSequence'].str.contains('!')].index]

        # load the tokenizer and pretrained protein model
        self.tokenizer = TAPETokenizer(vocab='iupac')
        self.model = ProteinBertForValuePrediction.from_pretrained(
            os.path.join(DATA_DIR, 'gfp_transformer_pretrained'))
        if self.use_cuda:
            self.model = self.model.cuda()

        # encode the entire dataset using the TAPE tokenizer
        x = np.array([self.tokenizer.encode(x.upper())
                      for x in df['aaSequence']])

        # convert the token ids to one-hot representations
        x = onehottify(x, n=30, dtype=np.float32)

        # format the fluorescence values to a tensor
        y = df['medianBrightness']\
            .to_numpy().astype(np.float32).reshape([-1, 1])

        # split the remaining proteins with a threshold
        ind = np.where(y <= np.percentile(
            y[:, 0], split_percentile))[0]

        # expose the designs
        x = x[ind]
        y = y[ind]

        mean_y = np.mean(y, axis=0, keepdims=True)
        st_y = np.std(y - mean_y, axis=0, keepdims=True)
        y = y + np.random.normal(0.0, 1.0, y.shape) * st_y * ys_noise

        # expose the designs
        self.x = x
        self.y = y

    def score(self,
              x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        scores = []
        with torch.no_grad():
            for i in range(x.shape[0] // self.batch_size):

                # run each batch through the torch model
                x_ids = torch.tensor(np.argmax(
                    x[i: (i + 1) * self.batch_size], axis=-1))
                if self.use_cuda:
                    x_ids = x_ids.cuda()
                y = self.model(x_ids)[0].cpu()
                scores.append(y.numpy().reshape([-1, 1]))

            if x.shape[0] % self.batch_size > 0:

                # if there are remaining elements run them at the end
                x_ids = torch.tensor(np.argmax(
                    x[-(x.shape[0] % self.batch_size):], axis=-1))
                if self.use_cuda:
                    x_ids = x_ids.cuda()
                y = self.model(x_ids)[0].cpu()
                scores.append(y.numpy().reshape([-1, 1]))

            # merge together all batches into a single numpy tensor
            return np.concatenate(scores, axis=0)



from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.conditional_task import ConditionalTask
from sklearn.ensemble import RandomForestRegressor
import numpy as np
import pickle as pkl
import os


def train_one_oracle(unique_c_i, x_i, y_i):
    """Train a Random Forest Regression Tree using scikit-learn
    and save that classifier to the disk

    Args:

    x: np.ndarray
        the training features for the decision tree represented as a
        matrix with a shape like [n_samples, n_features]
    c: np.ndarray
        an additional matrix of discrete ids that conditions the values y
        a matrix shaped like [n_samples]
    y: np.ndarray
        the training labels for the decision tree represented as a
        matrix with a shape like [n_samples, 1]
    """

    est = RandomForestRegressor(
        n_estimators=200,
        criterion="mse",
        max_depth=16,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.,
        max_features="auto",
        max_leaf_nodes=None,
        min_impurity_decrease=0.,
        min_impurity_split=None,
        bootstrap=True,
        oob_score=False,
        n_jobs=24,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0.0,
        max_samples=None)

    val_size = x_i.shape[0] // 4
    indices = np.arange(x_i.shape[0])
    np.random.shuffle(indices)
    x_train = x_i[indices[val_size:]]
    y_train = y_i[indices[val_size:]]
    x_test = x_i[indices[:val_size]]
    y_test = y_i[indices[:val_size]]

    est.fit(x_train, y_train[:, 0])
    with open(os.path.join(
            DATA_DIR, f'molecule_activity/rfr'
                      f'_{unique_c_i}.pkl'), 'wb') as f:
        pkl.dump(est, f)
    return est.score(x_test, y_test[:, 0])


def train_oracle(x, c, y):
    """Train a Random Forest Regression Tree using scikit-learn
    and save that classifier to the disk

    Args:

    x: np.ndarray
        the training features for the decision tree represented as a
        matrix with a shape like [n_samples, n_features]
    c: np.ndarray
        an additional matrix of discrete ids that conditions the values y
        a matrix shaped like [n_samples]
    y: np.ndarray
        the training labels for the decision tree represented as a
        matrix with a shape like [n_samples, 1]
    """

    r2 = []
    for unique_c_i in np.unique(c):
        indices = np.where(np.equal(c, unique_c_i))[0]
        r2.append(train_one_oracle(unique_c_i, x[indices], y[indices]))
    np.save(os.path.join(DATA_DIR, 'molecule_activity_rfr_r2.npy'), r2)


class MoleculeActivityTask(ConditionalTask):

    def score(self, x, c):
        return NotImplemented

    def __init__(self,
                 min_validation_r2=0.0,
                 min_num_samples=100,
                 ys_noise=0.0):
        """Create a task for designing super conducting materials that
        have a high critical temperature

        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        maybe_download('1_8c7uln7vzLbMmoviJhGWRqy-OyMZovc',
                       os.path.join(DATA_DIR, 'molecule_activity_X.npy'))
        maybe_download('1nZgMkKs6_hKb8o1Oy-KhGBLtI7HgqD2K',
                       os.path.join(DATA_DIR, 'molecule_activity_c.npy'))
        maybe_download('1vDycVRmfHi_-NiMjDaFrs-M4r9hZKr8T',
                       os.path.join(DATA_DIR, 'molecule_activity_y.npy'))
        maybe_download('1LAxrxhcsRhjE_p4J-Ft6TGd-AxeXGgyq',
                       os.path.join(DATA_DIR, 'molecule_activity_rfr_r2.npy'))
        maybe_download('17qbXsWQyaQ1jIQeOkMLC7GX6McIpgYGh',
                       os.path.join(DATA_DIR, 'molecule_activity.zip'))

        x = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_X.npy')).astype(np.float32)
        c = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_c.npy')).astype(np.int32)
        y = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_y.npy')).astype(np.float32)

        r2 = np.load(os.path.join(
            DATA_DIR, 'molecule_activity_rfr_r2.npy')).astype(np.float32)
        unique, counts = np.unique(c, return_counts=True)
        indices = np.where(np.logical_and(
            counts > min_num_samples, r2 > min_validation_r2))[0]

        self.x = []
        self.c = []
        self.y = []
        self.assays = []
        self.counts = []

        for idx in indices:
            unique_assay = unique[idx]
            matches = np.where(np.equal(c, unique_assay))[0]
            self.x.append(x[matches])
            self.c.append(c[matches])
            self.y.append(y[matches])
            self.assays.append([unique_assay])
            self.counts.append([counts[idx]])

        self.x = np.concatenate(self.x, axis=0)
        self.c = np.concatenate(self.c, axis=0)
        self.y = np.concatenate(self.y, axis=0)

        self.assays = np.concatenate(self.assays, axis=0)
        self.counts = np.concatenate(self.counts, axis=0)
        self.x = np.stack([1.0 - self.x, self.x], axis=2)

        mean_y = np.mean(self.y, axis=0, keepdims=True)
        st_y = np.std(self.y - mean_y, axis=0, keepdims=True)
        self.y = self.y + np.random.normal(
            0.0, 1.0, self.y.shape) * st_y * ys_noise

        self.oracles = []
        for assay_id in self.assays:
            with open(os.path.join(
                    DATA_DIR, f'molecule_activity/rfr'
                              f'_{assay_id}.pkl'), 'rb') as f:
                self.oracles.append(pkl.load(f))
        self.score = np.vectorize(
            self.scalar_score, signature='(n,2),()->(1)')

    def scalar_score(self,
                     x: np.ndarray,
                     c: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function
        c: np.ndarray
            a batch of task specification variables that determine
            how the designs are scores

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        est = self.oracles[int(np.where(np.equal(self.assays, c))[0])]
        return est.predict((x[np.newaxis, :, 1] > 0.5).astype(np.float32))



from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.task import Task
from sklearn.ensemble import GradientBoostingRegressor
import numpy as np
import pandas as pd
import pickle as pkl
import os


def train_oracle(x, y):
    """Train a Gradient Boosted Regression Tree using scikit-learn
    and save that classifier to the disk

    Args:

    x: np.ndarray
        the training features for the decision tree represented as a
        matrix with a shape like [n_samples, n_features]
    y: np.ndarray
        the training labels for the decision tree represented as a
        matrix with a shape like [n_samples, 1]
    """

    est = GradientBoostingRegressor(
        loss='ls',
        learning_rate=0.02,
        n_estimators=374,
        subsample=0.50,
        criterion='friedman_mse',
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_depth=16,
        min_impurity_decrease=0.0,
        min_impurity_split=None,
        init=None,
        random_state=None,
        max_features=None,
        alpha=0.9,
        verbose=0,
        max_leaf_nodes=None,
        warm_start=False,
        presort='deprecated',
        validation_fraction=1 / 3,
        n_iter_no_change=None,
        tol=0.0001,
        ccp_alpha=0.0)

    est.fit(x, y[:, 0])
    with open(os.path.join(
            DATA_DIR, 'superconductor_oracle.pkl'), 'wb') as f:
        pkl.dump(est, f)


class SuperconductorTask(Task):

    def __init__(self,
                 split_percentile=80,
                 ys_noise=0.0):
        """Create a task for designing super conducting materials that
        have a high critical temperature

        Args:

        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        maybe_download('1AguXqbNrSc665sablzVJh4RHLodeXglx',
                       os.path.join(DATA_DIR, 'superconductor_unique_m.csv'))
        maybe_download('15luLFnXpKDBi1jPL-NJlfeIGNI1QyZsf',
                       os.path.join(DATA_DIR, 'superconductor_train.csv'))
        maybe_download('1GvpMGXNuGVIoNgd0o7r-pXBQa1Zb-NSX',
                       os.path.join(DATA_DIR, 'superconductor_oracle.pkl'))

        train = pd.read_csv(os.path.join(DATA_DIR, 'superconductor_train.csv'))
        data = train.to_numpy()
        y = data[:, -1:]
        x = data[:, :-1]

        split_value = np.percentile(y[:, 0], split_percentile)
        indices = np.where(y <= split_value)[0]
        y = y[indices]
        x = x[indices]

        self.m = np.mean(x, axis=0, keepdims=True)
        self.st = np.std(x - self.m, axis=0, keepdims=True)

        mean_y = np.mean(y, axis=0, keepdims=True)
        st_y = np.std(y - mean_y, axis=0, keepdims=True)
        y = y + np.random.normal(0.0, 1.0, y.shape) * st_y * ys_noise

        self.y = y
        self.x = (x - self.m) / self.st

        with open(os.path.join(
                DATA_DIR, 'superconductor_oracle.pkl'), 'rb') as f:
            self.est = pkl.load(f)

    def score(self,
              x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        return self.est.predict(x * self.st + self.m).reshape([-1, 1])



from design_bench.registration import registry, register, make, spec
from design_bench.task import Task
import requests
import zipfile
import os


def maybe_download(fid,
                   destination,
                   unzip=True):
    """If a file does not already exist then download it from
    google drive using a custom GET request

    Args:

    fid: str
        the file id specified by google which is the 'X' in the url:
        https://drive.google.com/file/d/X/view?usp=sharing
    destination: str
        the destination for the file on this device; if the file
        already exists it will not be downloaded again
    """

    if not os.path.exists(destination):
        session = requests.Session()
        response = session.get(
            "https://docs.google.com/uc?export=download",
            params={'id': fid}, stream=True)
        token = get_confirm_token(response)
        if token is not None:
            response = session.get(
                "https://docs.google.com/uc?export=download",
                params={'id': fid, 'confirm': token}, stream=True)
        save_response(response, destination)
        if destination.endswith('.zip') and unzip:
            with zipfile.ZipFile(destination, 'r') as zip_ref:
                zip_ref.extractall(os.path.dirname(destination))


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value


def save_response(response, destination):
    os.makedirs(os.path.dirname(destination), exist_ok=True)
    with open(destination, "wb") as f:
        for chunk in response.iter_content(32768):
            if chunk:
                f.write(chunk)


DATA_DIR = os.path.join(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__))), 'data')

register(
    'Quadratic-v0',
    'design_bench.tasks.quadratic:QuadraticTask',
    kwargs=dict(
        global_optimum=(-1.0, 4.0, 2.0, -3.0, 5.0, 1.0,),
        oracle_noise_std=0.2,
        dataset_size=100,
        split_percentile=80,
        ys_noise=0.0))

register(
    'GFP-v0',
    'design_bench.tasks.gfp_v0:GFPV0Task',
    kwargs=dict(split_percentile=100,
                ys_noise=0.0))
register(
    'GFP-v1',
    'design_bench.tasks.gfp_v1:GFPV1Task',
    kwargs=dict(split_percentile=20,
                ys_noise=0.0))

register(
    'Superconductor-v0',
    'design_bench.tasks.superconductor:SuperconductorTask',
    kwargs=dict(split_percentile=80,
                ys_noise=0.0))

register(
    'MoleculeActivity-v0',
    'design_bench.tasks.molecule_activity_v0:MoleculeActivityV0Task',
    kwargs=dict(target_assay=600885,
                split_percentile=80,
                ys_noise=0.0))

try:

    import mujoco_py  # test that MuJoCo is installed
    import morphing_agents  # test that Morphing-Agents is installed
    from morphing_agents.mujoco.ant.env \
        import MorphingAntEnv
    from morphing_agents.mujoco.ant.elements \
        import LEG as ANT_LEG
    from morphing_agents.mujoco.ant.elements \
        import LEG_LOWER_BOUND as ANT_LEG_LOWER_BOUND
    from morphing_agents.mujoco.ant.elements \
        import LEG_UPPER_BOUND as ANT_LEG_UPPER_BOUND
    from morphing_agents.mujoco.dkitty.env \
        import MorphingDKittyEnv
    from morphing_agents.mujoco.dkitty.elements \
        import LEG as DKITTY_LEG
    from morphing_agents.mujoco.dkitty.elements \
        import LEG_LOWER_BOUND as DKITTY_LEG_LOWER_BOUND
    from morphing_agents.mujoco.dkitty.elements \
        import LEG_UPPER_BOUND as DKITTY_LEG_UPPER_BOUND

    register(
        'HopperController-v0',
        'design_bench.tasks.controller_v0:ControllerV0Task',
        kwargs=dict(
            obs_dim=11,
            action_dim=3,
            hidden_dim=64,
            env_name='Hopper-v2',
            x_file='hopper_controller_v0_X.npy',
            y_file='hopper_controller_v0_y.npy',
            split_percentile=100,
            ys_noise=0.0))

    register(
        'HopperController-v1',
        'design_bench.tasks.controller_v1:ControllerV1Task',
        kwargs=dict(
            obs_dim=11,
            action_dim=3,
            hidden_dim=256,
            env_name='Hopper-v2',
            x_file='hopper_controller_v1_X.npy',
            y_file='hopper_controller_v1_y.npy',
            split_percentile=80,
            ys_noise=0.0))

    register(
        'AntMorphology-v0',
        'design_bench.tasks.morphology_v0:MorphologyV0Task',
        kwargs=dict(
            env_class=MorphingAntEnv,
            elements=4,
            env_element=ANT_LEG,
            env_element_lb=ANT_LEG_LOWER_BOUND,
            env_element_ub=ANT_LEG_UPPER_BOUND,
            oracle_weights='ant_oracle.pkl',
            x_file='ant_morphology_X.npy',
            y_file='ant_morphology_y.npy',
            split_percentile=20,
            num_rollouts=1,
            rollout_horizon=100,
            num_parallel=1,
            ys_noise=0.0))

    register(
        'DKittyMorphology-v0',
        'design_bench.tasks.morphology_v0:MorphologyV0Task',
        kwargs=dict(
            env_class=MorphingDKittyEnv,
            elements=4,
            env_element=DKITTY_LEG,
            env_element_lb=DKITTY_LEG_LOWER_BOUND,
            env_element_ub=DKITTY_LEG_UPPER_BOUND,
            oracle_weights='dkitty_oracle.pkl',
            x_file='dkitty_morphology_X.npy',
            y_file='dkitty_morphology_y.npy',
            split_percentile=40,
            num_rollouts=1,
            rollout_horizon=100,
            num_parallel=1,
            ys_noise=0.0))

except ImportError as e:

    print('Skipping registration of: '
          'HopperController-v0, '
          'HopperController-v1, '
          'AntMorphology-v0, '
          'DKittyMorphology-v0')



from typing import Tuple
import numpy as np
import abc


class Task(abc.ABC):
    """A minimalistic task specification that includes designs
    represented by x and scores represented by y

    Attributes:

    x: np.ndarray
        the data set as a numpy tensor whose first 'axis' is
        the batch 'axis' and can take any shape
    y: np.ndarray
        the scores for every point in the data set represented
        as a tensor shaped like [num_samples, 1]
    input_shape: Tuple[int]
        Returns the shape of a single sample of x from the data set
        and excludes the batch 'axis'
    input_size: int
        Returns the size of a single sample of x from the data set
        and excludes the batch 'axis'

    Methods:

    score(self, x: np.ndarray) -> np.ndarray
        Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)
    """

    x: np.ndarray = None
    y: np.ndarray = None

    @abc.abstractmethod
    def score(self,
              x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        return NotImplemented

    @property
    def input_shape(self) -> Tuple[int]:
        """Returns the shape of a single sample of x from the data set
        and excludes the batch 'axis'

        Returns:

        shape: Tuple[int]
            a tuple of integers that corresponds to the shape of the
            inputs to a function of samples of x
        """

        return self.x.shape[1:]

    @property
    def input_size(self) -> int:
        """Returns the size of a single sample fo x from the data set
        and excludes the batch 'axis'

        Returns:

        size: int
            an integer that represents the total number of channels across
            all axes of a single sample from the data set
        """

        return int(np.prod(self.input_shape))



"""Code is primarily adapted from:
https://github.com/dhbrookes/CbAS/blob/master/util.py
"""


from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.task import Task
import pandas as pd
import numpy as np
import os


class GFPV0Task(Task):

    def __init__(self,
                 seed=0,
                 split_percentile=100,
                 ys_noise=0.0):
        """Load the GFP data set which includes maps from discrete
        protein designs to fluorescence scores

        Args:

        seed: int
            the random seed used for this experiment that determines the
            sampled values of data set noise
        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        maybe_download('1UO8L3uOp141m2v5dVlpGZ4tZ42XIJ4Vq',
                       os.path.join(DATA_DIR, 'gfp_gt_evals.npy'))
        maybe_download('1DeOoYQs5GEis3jIYsbGxuemjtsBiUSJm',
                       os.path.join(DATA_DIR, 'gfp_gpy.npy'))
        maybe_download('10xMOWXZjGOKLokO4jP6ya29-ZD2tb46X',
                       os.path.join(DATA_DIR, 'gfp_gpX.npy'))
        maybe_download('18EvOK25vmPvRGNbviv1Oep2CPXt3UrLt',
                       os.path.join(DATA_DIR, 'gfp_gpparams.npy'))
        maybe_download('1ySC8Rkfye6JfRKqoDS_KAXqUQTKtrbvZ',
                       os.path.join(DATA_DIR, 'gfp_gpKinv.npy'))
        maybe_download('1tRvY0W4ygoPxytdhAWZuwSQmvNj2QEtK',
                       os.path.join(DATA_DIR, 'gfp_gpK.npy'))
        maybe_download('1_jcPkQ-M1FRhkEONoE57WEbp_Rivkho2',
                       os.path.join(DATA_DIR, 'gfp_data.csv'))

        # load the static dataset
        self.sequence_gp = SequenceGP(load=True)
        x, _, y = get_experimental_X_y(random_state=seed)

        # cast everything to floats
        x = x.astype(np.float32)
        y = y.astype(np.float32).reshape([-1, 1])

        split_value = np.percentile(y[:, 0], split_percentile)
        indices = np.where(y <= split_value)[0]
        y = y[indices]
        x = x[indices]

        mean_y = np.mean(y, axis=0, keepdims=True)
        st_y = np.std(y - mean_y, axis=0, keepdims=True)
        y = y + np.random.normal(0.0, 1.0, y.shape) * st_y * ys_noise

        # expose the designs
        self.x = x
        self.y = y

    def score(self,
              x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        return self.sequence_gp.predict(np.argmax(
            x, axis=-1)).astype(np.float32).reshape([-1, 1])


class SequenceGP(object):

    def __init__(self,
                 load=False,
                 X_train=None,
                 y_train=None,
                 length_scale=1,
                 homo_noise=0.1,
                 load_prefix="gfp_gp",
                 k_beta=0.1,
                 c=1,
                 d=2):
        if load:
            self.load(prefix=load_prefix)
        else:
            assert X_train is not None and y_train is not None
            self.X_ = np.copy(X_train)
            self.y_ = np.copy(y_train).reshape((y_train.shape[0], 1))
            self.N_ = self.X_.shape[0]
            self.params_ = np.array([homo_noise, k_beta, c, d])
            self.K_ = None
            self.Kinv_ = None

    def _kernel(self,
                Xi,
                Xj):
        beta = self.params_[1]
        c = self.params_[2]
        d = self.params_[3]
        kij = np.prod(BLOSUM[(Xi, Xj)] ** beta)
        kii = np.prod(BLOSUM[(Xi, Xi)] ** beta)
        kjj = np.prod(BLOSUM[(Xj, Xj)] ** beta)
        k = kij / (np.sqrt(kii * kjj))
        k = np.exp(c * k)
        #         k = (k+c)**d
        return k

    def _fill_K(self,
                print_every=100):
        self.K_ = np.zeros((self.N_, self.N_))
        total = self.N_ * (self.N_ + 1) / 2
        m = 0
        homo_noise = self.params_[0]
        for i in range(self.N_):
            for j in range(i, self.N_):
                kij = self._kernel(self.X_[i], self.X_[j])
                if i == j:
                    kij += homo_noise
                self.K_[i, j] = kij
                self.K_[j, i] = kij

                m += 1
                if m % print_every == 0:
                    print("Number of K elements filled: %i / %i" % (m, total))

    def _invert_K(self):
        print("Inverting K...")
        self.Kinv_ = np.linalg.inv(self.K_)
        print("Done inverting K.")

    def build(self,
              print_every=100):
        self._fill_K(print_every=print_every)
        self._invert_K()

    def predict(self,
                Xstar,
                print_every=None,
                predict_variance=False):
        M = len(Xstar)
        Kstar = np.zeros((M, self.N_))
        total = M * self.N_
        m = 0
        for i in range(M):
            for j in range(self.N_):
                kij = self._kernel(Xstar[i], self.X_[j])
                Kstar[i, j] = kij
                m += 1
                if print_every is not None:
                    if m % print_every == 0:
                        print("Number of Kstar elements filled: %i / %i" % (m, total))
        mu_star = np.matmul(Kstar, np.matmul(self.Kinv_, self.y_))
        return mu_star

    def save(self,
             prefix="gfp_gp"):
        np.save(os.path.join(DATA_DIR, prefix + "X.npy", self.X_))
        np.save(os.path.join(DATA_DIR, prefix + "y.npy", self.y_))
        np.save(os.path.join(DATA_DIR, prefix + "K.npy", self.K_))
        np.save(os.path.join(DATA_DIR, prefix + "Kinv.npy", self.Kinv_))
        np.save(os.path.join(DATA_DIR, prefix + "params.npy", self.params_))

    def load(self,
             prefix="gfp_gp"):
        self.X_ = np.load(os.path.join(DATA_DIR, prefix + "X.npy"))
        self.y_ = np.load(os.path.join(DATA_DIR, prefix + "y.npy"))
        self.K_ = np.load(os.path.join(DATA_DIR, prefix + "K.npy"))
        self.Kinv_ = np.load(os.path.join(DATA_DIR, prefix + "Kinv.npy"))
        self.params_ = np.load(os.path.join(DATA_DIR, prefix + "params.npy"))
        self.N_ = self.X_.shape[0]


AA = ['a', 'r', 'n', 'd', 'c', 'q', 'e', 'g', 'h', 'i',
      'l', 'k', 'm', 'f', 'p', 's', 't', 'w', 'y', 'v']


AA_IDX = {AA[i]: i for i in range(len(AA))}


BLOSUM = np.array([
[3.9029,0.6127,0.5883,0.5446,0.8680,0.7568,0.7413,1.0569,0.5694,0.6325,0.6019,0.7754,0.7232,0.4649,0.7541,1.4721,0.9844,0.4165,0.5426,0.9365],
[0.6127,6.6656,0.8586,0.5732,0.3089,1.4058,0.9608,0.4500,0.9170,0.3548,0.4739,2.0768,0.6226,0.3807,0.4815,0.7672,0.6778,0.3951,0.5560,0.4201],
[0.5883,0.8586,7.0941,1.5539,0.3978,1.0006,0.9113,0.8637,1.2220,0.3279,0.3100,0.9398,0.4745,0.3543,0.4999,1.2315,0.9842,0.2778,0.4860,0.3690],
[0.5446,0.5732,1.5539,7.3979,0.3015,0.8971,1.6878,0.6343,0.6786,0.3390,0.2866,0.7841,0.3465,0.2990,0.5987,0.9135,0.6948,0.2321,0.3457,0.3365],
[0.8680,0.3089,0.3978,0.3015,19.5766,0.3658,0.2859,0.4204,0.3550,0.6535,0.6423,0.3491,0.6114,0.4390,0.3796,0.7384,0.7406,0.4500,0.4342,0.7558],
[0.7568,1.4058,1.0006,0.8971,0.3658,6.2444,1.9017,0.5386,1.1680,0.3829,0.4773,1.5543,0.8643,0.3340,0.6413,0.9656,0.7913,0.5094,0.6111,0.4668],
[0.7413,0.9608,0.9113,1.6878,0.2859,1.9017,5.4695,0.4813,0.9600,0.3305,0.3729,1.3083,0.5003,0.3307,0.6792,0.9504,0.7414,0.3743,0.4965,0.4289],
[1.0569,0.4500,0.8637,0.6343,0.4204,0.5386,0.4813,6.8763,0.4930,0.2750,0.2845,0.5889,0.3955,0.3406,0.4774,0.9036,0.5793,0.4217,0.3487,0.3370],
[0.5694,0.9170,1.2220,0.6786,0.3550,1.1680,0.9600,0.4930,13.5060,0.3263,0.3807,0.7789,0.5841,0.6520,0.4729,0.7367,0.5575,0.4441,1.7979,0.3394],
[0.6325,0.3548,0.3279,0.3390,0.6535,0.3829,0.3305,0.2750,0.3263,3.9979,1.6944,0.3964,1.4777,0.9458,0.3847,0.4432,0.7798,0.4089,0.6304,2.4175],
[0.6019,0.4739,0.3100,0.2866,0.6423,0.4773,0.3729,0.2845,0.3807,1.6944,3.7966,0.4283,1.9943,1.1546,0.3711,0.4289,0.6603,0.5680,0.6921,1.3142],
[0.7754,2.0768,0.9398,0.7841,0.3491,1.5543,1.3083,0.5889,0.7789,0.3964,0.4283,4.7643,0.6253,0.3440,0.7038,0.9319,0.7929,0.3589,0.5322,0.4565],
[0.7232,0.6226,0.4745,0.3465,0.6114,0.8643,0.5003,0.3955,0.5841,1.4777,1.9943,0.6253,6.4815,1.0044,0.4239,0.5986,0.7938,0.6103,0.7084,1.2689],
[0.4649,0.3807,0.3543,0.2990,0.4390,0.3340,0.3307,0.3406,0.6520,0.9458,1.1546,0.3440,1.0044,8.1288,0.2874,0.4400,0.4817,1.3744,2.7694,0.7451],
[0.7541,0.4815,0.4999,0.5987,0.3796,0.6413,0.6792,0.4774,0.4729,0.3847,0.3711,0.7038,0.4239,0.2874,12.8375,0.7555,0.6889,0.2818,0.3635,0.4431],
[1.4721,0.7672,1.2315,0.9135,0.7384,0.9656,0.9504,0.9036,0.7367,0.4432,0.4289,0.9319,0.5986,0.4400,0.7555,3.8428,1.6139,0.3853,0.5575,0.5652],
[0.9844,0.6778,0.9842,0.6948,0.7406,0.7913,0.7414,0.5793,0.5575,0.7798,0.6603,0.7929,0.7938,0.4817,0.6889,1.6139,4.8321,0.4309,0.5732,0.9809],
[0.4165,0.3951,0.2778,0.2321,0.4500,0.5094,0.3743,0.4217,0.4441,0.4089,0.5680,0.3589,0.6103,1.3744,0.2818,0.3853,0.4309,38.1078,2.1098,0.3745],
[0.5426,0.5560,0.4860,0.3457,0.4342,0.6111,0.4965,0.3487,1.7979,0.6304,0.6921,0.5322,0.7084,2.7694,0.3635,0.5575,0.5732,2.1098,9.8322,0.6580],
[0.9365,0.4201,0.3690,0.3365,0.7558,0.4668,0.4289,0.3370,0.3394,2.4175,1.3142,0.4565,1.2689,0.7451,0.4431,0.5652,0.9809,0.3745,0.6580,3.6922]]
)


def one_hot_encode_aa(aa_str,
                      pad=None):
    M = len(aa_str)
    aa_arr = np.zeros((M, 20), dtype=int)
    for i in range(M):
        aa_arr[i, AA_IDX[aa_str[i]]] = 1
    return aa_arr


def partition_data(X,
                   y,
                   percentile=40,
                   train_size=1000,
                   random_state=1,
                   return_test=False):
    np.random.seed(random_state)
    assert (percentile * 0.01 * len(y) >= train_size)
    y_percentile = np.percentile(y, percentile)
    idx = np.where(y < y_percentile)[0]
    #     print(y_percentile)
    rand_idx = np.random.choice(idx, size=train_size, replace=False)
    X_train = X[rand_idx]
    y_train = y[rand_idx]
    if return_test:
        test_idx = [i for i in idx if i not in rand_idx]
        X_test = X[test_idx]
        y_test = y[test_idx]
        return X_train, y_train, X_test, y_test
    else:
        return X_train, y_train


def get_experimental_X_y(random_state=1,
                         train_size=5000,
                         return_test=False,
                         return_all=False):
    df = pd.read_csv(os.path.join(DATA_DIR, 'gfp_data.csv'))
    X, _ = get_gfp_X_y_aa(df, large_only=True, ignore_stops=True)
    y_gt = np.load(os.path.join(DATA_DIR, "gfp_gt_evals.npy"))
    if return_test:
        X_train, gt_train, X_test, gt_test = partition_data(
            X, y_gt, percentile=20, train_size=train_size,
            random_state=random_state, return_test=return_test)
        np.random.seed(random_state)
        gt_var = 0.01
        y_train = gt_train + np.random.randn(*gt_train.shape) * gt_var
        y_test = gt_test + np.random.randn(*gt_test.shape) * gt_var
        return X_train, y_train, gt_train, X_test, y_test, gt_test
    else:
        X_train, gt_train = partition_data(
            X, y_gt, percentile=20, train_size=train_size,
            random_state=random_state, return_test=return_test)
        np.random.seed(random_state)
        gt_var = 0.01
        y_train = gt_train + np.random.randn(*gt_train.shape) * gt_var
        return X_train, y_train, gt_train


def get_gfp_X_y_aa(data_df,
                   large_only=False,
                   ignore_stops=True,
                   return_str=False):
    if large_only:
        idx = data_df.loc[(data_df['medianBrightness'] >
                           data_df['medianBrightness'].mean())].index
    else:
        idx = data_df.index
    data_df = data_df.loc[idx]

    if ignore_stops:
        idx = data_df.loc[~data_df['aaSequence'].str.contains('!')].index
    data_df = data_df.loc[idx]
    seqs = data_df['aaSequence']

    M = len(seqs[0])
    N = len(seqs)
    X = np.zeros((N, M, 20))
    j = 0
    for i in idx:
        X[j] = one_hot_encode_aa(seqs[i])
        j += 1
    y_raw = np.array(data_df['medianBrightness'][idx])
    y = y_raw
    if return_str:
        return X, y, list(data_df['aaSequence'])
    else:
        return X, y



from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.task import Task
import numpy as np
import gym
import os


class ControllerV1Task(Task):

    def score(self, x):
        return NotImplemented

    def __init__(self,
                 obs_dim=11,
                 action_dim=3,
                 hidden_dim=64,
                 env_name='Hopper-v2',
                 x_file='hopper_controller_v1_X.npy',
                 y_file='hopper_controller_v1_y.npy',
                 split_percentile=100,
                 ys_noise=0.0):
        """Load static datasets of weights and their corresponding
        expected returns from the disk

        Args:

        obs_dim: int
            the number of channels in the environment observations
        action_dim: int
            the number of channels in the environment actions
        hidden_dim: int
            the number of channels in policy hidden layers
        env_name: str
            the name of the gym.Env to use when collecting rollouts
        x_file: str
            the name of the dataset file to be loaded for x
        y_file: str
            the name of the dataset file to be loaded for y
        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        maybe_download('14AdCoQT0F4YSOjJCtw3-CqlI18HyN3t1',
                       os.path.join(DATA_DIR, 'hopper_controller_v1_X.npy'))
        maybe_download('1uEiVCt2Da7BdMNQVk9gB13qra1ZPJt3w',
                       os.path.join(DATA_DIR, 'hopper_controller_v1_y.npy'))

        self.obs_dim = obs_dim
        self.action_dim = action_dim
        self.hidden_dim = hidden_dim
        self.env_name = env_name

        x = np.load(os.path.join(DATA_DIR, x_file))
        y = np.load(os.path.join(DATA_DIR, y_file))
        x = x.astype(np.float32)
        y = y.astype(np.float32).reshape([-1, 1])

        split_value = np.percentile(y[:, 0], split_percentile)
        indices = np.where(y <= split_value)[0]
        y = y[indices]
        x = x[indices]

        mean_y = np.mean(y, axis=0, keepdims=True)
        st_y = np.std(y - mean_y, axis=0, keepdims=True)
        y = y + np.random.normal(0.0, 1.0, y.shape) * st_y * ys_noise

        self.x = x
        self.y = y
        self.score = np.vectorize(self.scalar_score,
                                  signature='(n)->(1)')


    @property
    def stream_shapes(self):
        return ((self.hidden_dim, self.obs_dim),
                (self.hidden_dim,),
                (self.action_dim, self.hidden_dim),
                (self.action_dim,),
                (1,),
                (1,))

    @property
    def stream_sizes(self):
        return [self.obs_dim * self.hidden_dim,
                self.hidden_dim,
                self.hidden_dim * self.action_dim,
                self.action_dim,
                1,
                1]

    def scalar_score(self,
                     x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        # extract weights from the vector design
        weights = []
        for s in self.stream_shapes:
            weights.append(x[0:np.prod(s)].reshape(s))
            x = x[np.prod(s):]

        # the final two weights are for log_std and not used
        weights.pop(-1)
        weights.pop(-1)

        # create a policy forward pass in numpy
        def mlp_policy(h):
            h = np.maximum(0, h @ weights[0].T + weights[1])
            return h @ weights[2].T + weights[3]

        # make a copy of the policy and the environment
        env = gym.make(self.env_name)

        # perform a single rollout for quick evaluation
        obs, done = env.reset(), False
        path_returns = np.zeros([1], dtype=np.float32)
        while not done:
            obs, rew, done, info = env.step(mlp_policy(obs))
            path_returns += rew.astype(np.float32)
        return path_returns



from setuptools import find_packages
from setuptools import setup


F = 'README.md'
with open(F, 'r') as readme:
    LONG_DESCRIPTION = readme.read()


INSTALL_REQUIRES = ['numpy', 'pandas', 'requests', 'scikit-learn', 'tape_proteins']
EXTRA_REQUIRES = ['gym[mujoco]', 'morphing-agents']


CLASSIFIERS = [
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8']


setup(
    name='design-bench',
    packages=find_packages(include=['design_bench', 'design_bench.*']),
    version='1.4',
    license='MIT',
    description='Design-Bench: Benchmarks for Data-Driven Offline Model-Based Optimization',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Brandon Trabucco',
    author_email='brandon@btrabucco.com',
    url='https://github.com/brandontrabucco/design-bench',
    download_url='https://github.com/brandontrabucco/design-bench/archive/v1_4.tar.gz',
    keywords=['Offline', 'Benchmark', 'Model-Based Optimization'],
    install_requires=INSTALL_REQUIRES,
    extras_require={'all': EXTRA_REQUIRES},
    classifiers=CLASSIFIERS)



from design_bench import DATA_DIR
from design_bench import maybe_download
from design_bench.task import Task
from morphing_agents.mujoco.ant.env import MorphingAntEnv
from morphing_agents.mujoco.ant.elements import LEG
from morphing_agents.mujoco.ant.elements import LEG_LOWER_BOUND
from morphing_agents.mujoco.ant.elements import LEG_UPPER_BOUND
from multiprocessing import Pool
import numpy as np
import os
import pickle as pkl


class MorphologyV0Task(Task):

    def __init__(self,
                 env_class=MorphingAntEnv,
                 elements=4,
                 env_element=LEG,
                 env_element_lb=LEG_LOWER_BOUND,
                 env_element_ub=LEG_UPPER_BOUND,
                 oracle_weights='ant_oracle.pkl',
                 x_file='ant_morphology_X.npy',
                 y_file='ant_morphology_y.npy',
                 split_percentile=60,
                 num_rollouts=1,
                 rollout_horizon=100,
                 num_parallel=1,
                 ys_noise=0.0):
        """Load static datasets of weights and their corresponding
        expected returns from the disk

        Args:

        num_parallel: int
            the number of parallel trials in the docker container
        num_gpus: int
            the number of gpus to use in this docker container
        n_envs: int
            the number of parallel sampling environments for PPO
        max_episode_steps: int
            the maximum length of a episode when collecting samples
        total_timesteps: int
            the maximum number of samples to collect from the environment
        domain: str
            the particular morphology domain such as 'ant' or 'dog'
        split_percentile: int
            the percentile (out of 100) to split the data set by and only
            include samples with score below this percentile
        ys_noise: float
            the number of standard deviations of noise to add to
            the static training dataset y values accompanying this task
        """

        self.pool = Pool(num_parallel)
        maybe_download('12H-4AvzpMVmq7M7b7nD_RPu5GNeCuCbu',
                       os.path.join(DATA_DIR, 'ant_morphology_X.npy'))
        maybe_download('1uSF6oc7OlLGioe_sZQwmjibuRbPCYRGu',
                       os.path.join(DATA_DIR, 'ant_morphology_y.npy'))
        maybe_download('1FJY6LQG3kvLIx00WT2XVGyfU9JrmO0T0',
                       os.path.join(DATA_DIR, 'ant_oracle.pkl'))
        maybe_download('17PqWmmbIiUwNBxLwhjTnV38Dx2ejefCQ',
                       os.path.join(DATA_DIR, 'dkitty_morphology_X.npy'))
        maybe_download('1DggTSnRn_SzmYonbLgLppzWC8OtqYAb8',
                       os.path.join(DATA_DIR, 'dkitty_morphology_y.npy'))
        maybe_download('1GpZc9UezjyvAn_ejwG3cibq1stQMjiiM',
                       os.path.join(DATA_DIR, 'dkitty_oracle.pkl'))

        self.env_class = env_class
        self.elements = elements
        self.env_element = env_element
        self.lb = env_element_lb
        self.ub = env_element_ub

        self.num_rollouts = num_rollouts
        self.rollout_horizon = rollout_horizon
        with open(os.path.join(
                DATA_DIR, oracle_weights), 'rb') as f:
            self.weights = pkl.load(f)

        x = np.load(os.path.join(DATA_DIR, x_file))
        y = np.load(os.path.join(DATA_DIR, y_file))
        x = x.astype(np.float32)
        y = y.astype(np.float32).reshape([-1, 1])

        # remove all samples above the qth percentile in the data set
        split_temp = np.percentile(y[:, 0], split_percentile)
        indices = np.where(y <= split_temp)[0]
        self.x = x[indices].astype(np.float32)
        self.y = y[indices].astype(np.float32)

        mean_y = np.mean(self.y, axis=0, keepdims=True)
        st_y = np.std(self.y - mean_y, axis=0, keepdims=True)
        self.y = self.y + np.random.normal(
            0.0, 1.0, self.y.shape) * st_y * ys_noise

    def score(self,
              x: np.ndarray) -> np.ndarray:
        """Calculates a score for the provided tensor x using a ground truth
        oracle function (the goal of the task is to maximize this)

        Args:

        x: np.ndarray
            a batch of sampled designs that will be evaluated by
            an oracle score function

        Returns:

        scores: np.ndarray
            a batch of scores that correspond to the x values provided
            in the function argument
        """

        return np.stack(
            self.pool.map(scalar_score, [
                (self.weights,
                 self.env_class,
                 self.env_element,
                 self.ub,
                 self.lb,
                 self.elements,
                 self.num_rollouts,
                 self.rollout_horizon,
                 xi) for xi in x]), axis=0)


def scalar_score(args):
    """Calculates a score for the provided tensor x using a ground truth
    oracle function (the goal of the task is to maximize this)

    Args:

    x: np.ndarray
        a batch of sampled designs that will be evaluated by
        an oracle score function

    Returns:

    scores: np.ndarray
        a batch of scores that correspond to the x values provided
        in the function argument
    """

    # unpack the list of args
    (weights,
     env_class,
     env_element,
     ub,
     lb,
     elements,
     num_rollouts,
     rollout_horizon,
     x) = args

    # create a policy forward pass in numpy
    def mlp_policy(h):
        h = np.maximum(0.0, h @ weights[0] + weights[1])
        h = np.maximum(0.0, h @ weights[2] + weights[3])
        return np.tanh(np.split(
            h @ weights[4] + weights[5], 2)[0])

    # convert vectors to morphologies
    env = env_class(expose_design=False, fixed_design=[
        env_element(*np.clip(np.array(xi), lb, ub))
        for xi in np.split(x, elements)])

    # do many rollouts using a pretrained agent
    average_returns = []
    for i in range(num_rollouts):
        obs = env.reset()
        average_returns.append(np.zeros([], dtype=np.float32))
        for t in range(rollout_horizon):
            obs, rew, done, info = env.step(mlp_policy(obs))
            average_returns[-1] += rew.astype(np.float32)
            if done:
                break

    # we average here so as to reduce randomness
    return np.mean(
        average_returns).reshape([1])
